import base64
import json
import logging
import os

from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError

from gradient_utils.exceptions import ConfigError

logger = logging.getLogger(__name__)


def _get_work_env():
    env = os.environ.get("WORK_ENV")

    if not env:
        env = "local"

    return env


def _mongo_db_host():
    return os.environ.get("MONGO_DB_HOST")


def _mongo_db_port():
    return os.environ.get("MONGO_DB_PORT")


def _experiment_name():
    return os.environ.get("EXPERIMENT_NAME")


def _get_paperspace_tf_config():
    tf_config = os.environ.get("PS_CONFIG")

    if not tf_config:
        env = _get_work_env()
        if env != "local":
            return
        else:
            raise ConfigError(
                component="TF Config",
                message="Something went wrong. Check if you set PS_CONFIG"
            )
    if tf_config:
        return json.loads(base64.urlsafe_b64decode(tf_config).decode("utf-8"))


def _check_mongo_client_connection():
    """
    Function to check if connection to mongo is possible.

    :return: bool value if ping to mongo db ended with success
    """
    mongo_status = False

    mongo_port = _mongo_db_port()
    if mongo_port:
        mongo_port = int(mongo_port)

    client = MongoClient(
        host=_mongo_db_host(),
        port=mongo_port
    )
    try:
        resp = client.db_name.command('ping')
    except ServerSelectionTimeoutError:
        logger.warning("Check mongo db connection - connection to mongo db timeout")
        return mongo_status

    if resp.get("ok"):
        mongo_status = True

    return mongo_status


def get_mongo_conn_str():
    """
    Function to check and construct mongo db connection string.

    :raise: ConfigError with information about missing values.

    :return: Mongo connection string
    """
    mongo_host = _mongo_db_host()
    mongo_port = _mongo_db_port()
    experiment_name = _experiment_name()

    if mongo_host and mongo_port and experiment_name:
        return "mongo://%s:%s/%s/jobs" % (mongo_host, mongo_port, experiment_name)
    else:
        error_message = "Something went wrong. " \
                        "Check os variables that are needed for constricting mongo connection string: " \
                        "- MONGO_DB_HOST: %s  " \
                        "- MONGO_DB_PORT: %s " \
                        "- EXPERIMENT_NAME: %s" % (mongo_host, mongo_port, experiment_name)
        raise ConfigError(
            component="Mongo connection",
            message=error_message
        )


def data_dir():
    return os.environ.get("PS_JOBSPACE", "/storage")


def model_dir(model_name=None):
    if model_name:
        model_path = os.path.join(os.environ.get("PS_MODEL_PATH", "/storage/models/"), model_name)
    else:
        model_path = os.environ.get("PS_MODEL_PATH", "/storage/models/")
    return model_path


def export_dir(model_name=None):
    if model_name:
        export_path = os.path.join(os.environ.get("PS_MODEL_PATH", "/storage/models/"), model_name)
    else:
        export_path = os.environ.get("PS_MODEL_PATH", "/storage/models/")
    return export_path


def worker_hosts():
    return os.environ.get("WORKER_HOSTS")


def ps_hosts():
    return os.environ.get("PS_HOSTS")


def task_index():
    return os.environ.get("INDEX")


def job_name():
    return os.environ.get("TYPE")



"""
Gradient ML SDK
"""
from gradient_utils.hyper_parameter import hyper_tune
from gradient_utils.metrics import MetricsLogger
from gradient_utils.multi_node import get_tf_config
from gradient_utils.utils import get_mongo_conn_str, data_dir, worker_hosts, export_dir, job_name, model_dir, ps_hosts, \
    task_index

__version__ = "0.1.0"

__all__ = [
    "__version__",
    "get_tf_config",
    "get_mongo_conn_str",
    "data_dir",
    "worker_hosts",
    "export_dir",
    "job_name",
    "model_dir",
    "ps_hosts",
    "task_index",
    "hyper_tune",
    "MetricsLogger",
]



import json
import os

from gradient_utils.exceptions import ConfigError
from gradient_utils.utils import _get_paperspace_tf_config


def get_tf_config():
    """
    Function to prepare TensorFlow config and set os env with proper configuration

    :raise: ConfigError when there is no configuration for TensorFlow to set as os env
    """
    tf_config = _get_paperspace_tf_config()
    if tf_config:
        os.environ['TF_CONFIG'] = json.dumps(tf_config)
    else:
        raise ConfigError(
            component="TF Config",
            message="Something went wrong. For some reason there is no configuration for TensorFlow."
        )



import os
from typing import Union

from prometheus_client import push_to_gateway, Gauge, CollectorRegistry, Counter, Summary, Histogram, Info


def get_metric_pushgateway():
    return os.getenv('PAPERSPACE_METRIC_PUSHGATEWAY', 'http://localhost:9091')


def _get_env_var_or_raise(*env_vars):
    rv = None
    for env_var in env_vars:
        rv = os.getenv(env_var)
        if not rv:
            break

    if rv is None:
        msg = "{} environment variable(s) not found".format(", ".join(env_vars))
        raise ValueError(msg)

    return rv


def get_job_id():
    return _get_env_var_or_raise("PAPERSPACE_EXPERIMENT_ID", "PAPERSPACE_DEPLOYMENT_ID")


class MetricsLogger:
    """Prometheus wrapper for logging custom metrics
    
    Examples:
        >>> from gradient_utils import MetricsLogger
        >>> m_logger = MetricsLogger()
        >>> m_logger.add_gauge("some_metric_1")
        >>> m_logger.add_gauge("some_metric_2")
        >>> m_logger["some_metric_1"].set(3)
        >>> m_logger["some_metric_1"].inc()
        >>> m_logger["some_metric_2"].set_to_current_time()
        >>> m_logger.push_metrics()

    """

    def __init__(self, job_id=None, registry=CollectorRegistry(), grouping_key: dict = None, push_gateway=None):
        self.id = job_id or get_job_id()
        self.registry = registry
        self.grouping_key = grouping_key
        self.push_gateway = push_gateway or get_metric_pushgateway()

        self._metrics = dict()

    def __getitem__(self, item: str) -> Union[Gauge, Counter, Summary, Histogram, Info]:
        return self._metrics[item]

    def add_gauge(self, name: str):
        self._add_metric(Gauge, name)

    def add_counter(self, name: str):
        self._add_metric(Counter, name)

    def add_summary(self, name: str):
        self._add_metric(Summary, name)

    def add_histogram(self, name: str):
        self._add_metric(Histogram, name)

    def add_info(self, name: str):
        self._add_metric(Info, name)

    def _add_metric(self, cls, name, documentation=""):
        new_metric = cls(name, documentation=documentation, registry=self.registry)
        self._metrics[name] = new_metric

    def push_metrics(self, timeout=30):
        push_to_gateway(
            gateway=self.push_gateway,
            job=self.id,
            registry=self.registry,
            grouping_key=self.grouping_key,
            timeout=timeout,
        )



import logging
import os
from time import sleep

from hyperopt import fmin, tpe
from hyperopt.mongoexp import MongoTrials
from pymongo.errors import ServerSelectionTimeoutError

from gradient_utils.utils import get_mongo_conn_str, _experiment_name, _check_mongo_client_connection

logger = logging.getLogger(__name__)


def _hyper_tune_check():
    """
    Function to check all requirements for hyper tune

    :return: bool value when all check complete
    """
    max_retry = os.environ.get("MAX_CONNECTION_RETRY", 5)
    retry_timeout = os.environ.get("CONNECTION_RETRY_TIMEOUT", 5)

    is_check_complete = False

    for i in range(max_retry):
        logger.warning("Hyper Tune Check - check connection to mongo db")
        is_connection_available = _check_mongo_client_connection()

        if not is_connection_available:
            current_retry_timeout = (i + 1) * retry_timeout
            logger.warning(
                "Hyper Tune Check - connection is not available retry next connection check in %s seconds",
                current_retry_timeout
            )
            sleep(current_retry_timeout)
        else:
            is_check_complete = True
            break

    return is_check_complete


def hyper_tune(train_model, hparam_def, algo=tpe.suggest, max_evals=25, func=fmin):
    """
    Function to prepare and run hyper parameter tune.

    :param train_model: User model to tune
    :param hparam_def: User hyper tune param definition
    :param algo: Search algorithm
    :param max_evals: Allow up to this many function evaluations before returning
    :param func: function that will run hyper tune logic, by default hyperopt fmin function

    :return: None if there is no result or dict with result for tune
    """
    mongo_connect_str = get_mongo_conn_str()
    is_connection_available = True

    while is_connection_available:
        try:
            trials = MongoTrials(mongo_connect_str, exp_key=_experiment_name())
        except ServerSelectionTimeoutError:
            logger.warning("Hyper Tune - MongoTrials server selection Timeout Error")
            is_connection_available = _hyper_tune_check()
        else:
            return func(train_model, space=hparam_def, trials=trials, algo=algo, max_evals=max_evals)



import re
from codecs import open
from os import path

from setuptools import setup, find_packages

with open("gradient_utils/__init__.py", "r", encoding="utf8") as f:
    version = re.search(r"__version__ = \"(.*?)\"", f.read()).group(1)

here = path.abspath(path.dirname(__file__))


# Get the long description from the README file
with open("README.md", "r") as fh:
    long_description = fh.read()


requirements = [
    "wheel",
    "hyperopt<0.2",  # hyperopt==0.2 does not work on Py<3.6
    "pymongo",
    "prometheus_client",
]

dev_requirements = [
    "tox"
]

setup(
    name="gradient_utils",
    version=version,
    description="Gradient Utils",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Paperspace/gradient-utils",
    author="Paperspace Co.",
    author_email="info@paperspace.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="gradient utils sdk",
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    install_requires=requirements,
    extras_require={
        "dev": dev_requirements
    }
)



class ConfigError(Exception):
    def __init__(self, component, message):
        self.component = component
        self.message = message

    def __str__(self):
        return "For component - %s - received error message: %s" % (self.component, self.message)
